@@ -1,6 +1,6 @@
 from __future__ import annotations
 
-import ast
+from collections import defaultdict
 from string import Template
 from typing import TYPE_CHECKING, Any, List, Optional, Union, cast
 
@@ -34,8 +34,8 @@ from robotcode.language_server.robotframework.utils.ast_utils import (
     range_from_node,
     range_from_token,
 )
-from robotcode.language_server.robotframework.utils.async_ast import Visitor
 
+from .code_action_helper_mixin import CodeActionHelperMixin, FindSectionsVisitor
 from .model_helper import ModelHelperMixin
 from .protocol_part import RobotLanguageServerProtocolPart
 
@@ -61,41 +61,7 @@ ${name}
 )
 
 
-class FindSectionsVisitor(Visitor):
-    def __init__(self) -> None:
-        self.keyword_sections: List[ast.AST] = []
-        self.variable_sections: List[ast.AST] = []
-        self.setting_sections: List[ast.AST] = []
-        self.testcase_sections: List[ast.AST] = []
-        self.sections: List[ast.AST] = []
-
-    def visit_KeywordSection(self, node: ast.AST) -> None:  # noqa: N802
-        self.keyword_sections.append(node)
-        self.sections.append(node)
-
-    def visit_VariableSection(self, node: ast.AST) -> None:  # noqa: N802
-        self.variable_sections.append(node)
-        self.sections.append(node)
-
-    def visit_SettingSection(self, node: ast.AST) -> None:  # noqa: N802
-        self.setting_sections.append(node)
-        self.sections.append(node)
-
-    def visit_TestCaseSection(self, node: ast.AST) -> None:  # noqa: N802
-        self.testcase_sections.append(node)
-        self.sections.append(node)
-
-    def visit_CommentSection(self, node: ast.AST) -> None:  # noqa: N802
-        self.sections.append(node)
-
-
-def find_keyword_sections(node: ast.AST) -> Optional[List[ast.AST]]:
-    visitor = FindSectionsVisitor()
-    visitor.visit(node)
-    return visitor.keyword_sections if visitor.keyword_sections else None
-
-
-class RobotCodeActionQuickFixesProtocolPart(RobotLanguageServerProtocolPart, ModelHelperMixin):
+class RobotCodeActionQuickFixesProtocolPart(RobotLanguageServerProtocolPart, ModelHelperMixin, CodeActionHelperMixin):
     _logger = LoggingDescriptor()
 
     def __init__(self, parent: RobotLanguageServerProtocol) -> None:
@@ -260,44 +226,9 @@ class RobotCodeActionQuickFixesProtocolPart(RobotLanguageServerProtocolPart, Mod
         model = await self.parent.documents_cache.get_model(document, False)
         namespace = await self.parent.documents_cache.get_namespace(document)
 
-        keyword_sections = find_keyword_sections(model)
-        keyword_section = keyword_sections[-1] if keyword_sections else None
-
-        lines = document.get_lines()
-
-        if keyword_section is not None:
-            node_range = range_from_node(keyword_section, skip_non_data=True, allow_comments=True)
-            insert_pos = Position(node_range.end.line + 1, 0)
-            insert_range = Range(insert_pos, insert_pos)
-            insert_text = f"\n\n{insert_text}"
-        else:
-            if namespace.languages is None or not namespace.languages.languages:
-                keywords_text = "Keywords"
-            else:
-                keywords_text = namespace.languages.languages[-1].keywords_header
-
-            insert_text = f"\n\n*** {keywords_text} ***\n{insert_text}"
-
-            end_line = len(lines) - 1
-            while end_line >= 0 and not lines[end_line].strip():
-                end_line -= 1
-            doc_pos = Position(end_line + 1, 0)
-
-            insert_range = Range(doc_pos, doc_pos)
-
-        if insert_range.start.line >= len(lines) and lines[-1].strip():
-            doc_pos = Position(len(lines) - 1, len(lines[-1]))
-            insert_range = Range(doc_pos, doc_pos)
-            insert_text = "\n" + insert_text
-
-        if insert_range.start.line <= len(lines) and lines[insert_range.start.line].startswith("*"):
-            insert_text = insert_text + "\n\n"
-        if (
-            insert_range.start.line + 1 <= len(lines)
-            and not lines[insert_range.start.line].strip()
-            and lines[insert_range.start.line + 1].startswith("*")
-        ):
-            insert_text = insert_text + "\n"
+        insert_text, insert_range = await self.create_insert_keyword_workspace_edit(
+            document, model, namespace, insert_text
+        )
 
         we = WorkspaceEdit(
             document_changes=[
@@ -320,24 +251,36 @@ class RobotCodeActionQuickFixesProtocolPart(RobotLanguageServerProtocolPart, Mod
     async def code_action_disable_robotcode_diagnostics_for_line(
         self, document: TextDocument, range: Range, context: CodeActionContext
     ) -> Optional[List[Union[Command, CodeAction]]]:
-        if (context.only and CodeActionKind.QUICK_FIX in context.only) or context.trigger_kind in [
-            CodeActionTriggerKind.INVOKED,
-            CodeActionTriggerKind.AUTOMATIC,
-        ]:
+        if (
+            range.start.line == range.end.line
+            and range.start.character <= range.end.character
+            and (
+                (context.only and CodeActionKind.QUICK_FIX in context.only)
+                or context.trigger_kind
+                in [
+                    CodeActionTriggerKind.INVOKED,
+                    CodeActionTriggerKind.AUTOMATIC,
+                ]
+            )
+        ):
             all_diagnostics = [d for d in context.diagnostics if d.source and d.source.startswith("robotcode.")]
             if all_diagnostics:
+                result = defaultdict(list)
+                for diagnostics in all_diagnostics:
+                    result[diagnostics.code].append(diagnostics)
+
                 return [
                     CodeAction(
-                        f"Disable '{diagnostics.code}' for this line",
+                        f"Disable '{k}' for this line",
                         kind=CodeActionKind.QUICK_FIX,
                         command=Command(
                             self.parent.commands.get_command_name(self.disable_robotcode_diagnostics_for_line_command),
                             self.parent.commands.get_command_name(self.disable_robotcode_diagnostics_for_line_command),
                             [document.document_uri, range],
                         ),
-                        diagnostics=[diagnostics],
+                        diagnostics=v,
                     )
-                    for diagnostics in all_diagnostics
+                    for k, v in result.items()
                 ]
 
         return None
